from __future__ import annotations

from .base import AbstractTimer


class PeriodicTimer(AbstractTimer):
    every: int | float
    strict: bool

    def __init__(self, every: int | float, strict=False):
        super().__init__()
        self.every = every
        self.strict = strict

    def next_due(self, last_due: float, now: float) -> int | float:
        """
        Return unix time for the next time the trigger is due
        """
        if self.strict:
            next = last_due + self.every

        else:
            next = last_due

        if next <= now:
            next = now + self.every

        return next



"""
HTTP server

Wrapper around aiohttp, https://pypi.org/project/aiohttp/
"""

from __future__ import annotations

from asyncio import sleep
from typing import TYPE_CHECKING

from aiohttp import web

from ..constants import DEFAULT_HOST, DEFAULT_PORT
from ..status import Status
from .base import AbstractAsyncioServer, AbstractServer
from .connections.http import WebSocketConnection

if TYPE_CHECKING:
    from ..app import App


class WebSocketServer(AbstractServer):
    """
    A websocket server is a sub-server of an HttpServer.

    To create a websocket server, call HttpServer.create_websocket(..)
    """

    connection_class = WebSocketConnection

    def __init__(self, http_server: HttpServer):
        self.http_server = http_server
        super().__init__()

    async def listen_loop(self):
        # Keep the server active
        while self.http_server.status != Status.STOPPED:
            await sleep(0.1)

    async def handle_connect(self, request):
        connection: WebSocketConnection = self.connection_class(server=self)
        await connection.prepare(request)
        await self.connected(connection)

        # Keep the connection open
        while connection.connected:
            await sleep(0.1)


class HttpServer(AbstractAsyncioServer):
    #: Host
    host: str

    #: Port
    port: int

    #: aiohttp web.Application instance
    web_app: web.Application

    #: websocket servers provided by this server
    websockets: list[WebSocketServer]

    def __init__(
        self,
        host: str = DEFAULT_HOST,
        port: int = DEFAULT_PORT,
        web_app: web.Application | None = None,
        **handler_kwargs,
    ):
        self.host = host
        self.port = port
        if web_app:
            self.web_app = web_app
        else:
            self.web_app = web.Application()
        self.handler_kwargs = handler_kwargs

        self.websockets: list[WebSocketServer] = []
        super().__init__()

    def __str__(self):
        return f"HTTP {self.host}:{self.port}"

    async def run(self, app: App):
        for ws in self.websockets:
            app.add_server(ws)
        await super().run(app)

    async def create(self):
        await super().create()

        loop = self.app.loop if self.app else None
        if loop is None:
            raise ValueError("Cannot start HttpServer without running loop")

        self.server = await loop.create_server(
            protocol_factory=self.web_app.make_handler(**self.handler_kwargs),
            host=self.host,
            port=self.port,
        )

    def add_route(self, method: str, path: str, *, name=None, expect_handler=None):
        def decorator(handler):
            self.web_app.router.add_route(
                method, path, handler, name=name, expect_handler=expect_handler
            )
            return handler

        return decorator

    def add_get(self, path: str, **kwargs):
        return self.add_route("GET", path, **kwargs)

    def add_post(self, path: str, **kwargs):
        return self.add_route("POST", path, **kwargs)

    def add_head(self, path: str, **kwargs):
        return self.add_route("HEAD", path, **kwargs)

    def add_put(self, path: str, **kwargs):
        return self.add_route("PUT", path, **kwargs)

    def add_patch(self, path: str, **kwargs):
        return self.add_route("PATCH", path, **kwargs)

    def add_delete(self, path: str, **kwargs):
        return self.add_route("DELETE", path, **kwargs)

    def add_view(self, path: str, **kwargs):
        return self.add_route("*", path, **kwargs)

    def create_websocket(self, path: str):
        ws = WebSocketServer(self)
        self.add_get(path)(ws.handle_connect)
        self.websockets.append(ws)
        if self.app:
            self.app.add_server(ws)
        return ws



"""
Connection events
"""

from __future__ import annotations

from .base import Event

__all__ = ["Connection", "Connect", "Receive", "Disconnect"]


class Connection(Event):
    "Connection event"

    def __init__(self, connection):
        super(Connection, self).__init__()
        self.connection = connection
        self.server = getattr(connection, "server", None)

    def __str__(self):
        msg = super(Connection, self).__str__().strip()
        return f"{msg} ({self.connection})"


class Connect(Connection):
    "Connection connected"


class Disconnect(Connection):
    "Connection disconnected"


class Receive(Connection):
    "Data received"

    def __init__(self, connection, data):
        super(Receive, self).__init__(connection)
        self.data = data

    def __str__(self):
        msg = super(Receive, self).__str__().strip()
        return f"{msg}: {self.data}"



"""
Telnet server

Wrapper around telnetlib3, https://pypi.org/project/telnetlib3/
"""

from __future__ import annotations

from ..constants import DEFAULT_HOST, DEFAULT_PORT
from .base import AbstractAsyncioServer
from .connections.telnet import TelnetConnection

try:
    import telnetlib3
except ImportError:
    raise ValueError("telnetlib3 not found - pip install mara[telnet]")


class TelnetServer(AbstractAsyncioServer):
    connection_class: type[TelnetConnection] = TelnetConnection

    def __init__(
        self, host: str = DEFAULT_HOST, port: int = DEFAULT_PORT, **telnet_kwargs
    ):
        self.host = host
        self.port = port

        if "shell" in telnet_kwargs:
            raise ValueError("Cannot specify a shell for TelnetServer")
        self.telnet_kwargs = telnet_kwargs
        self.telnet_kwargs["shell"] = self.handle_connect

        super().__init__()

    def __str__(self):
        return f"Telnet {self.host}:{self.port}"

    async def create(self):
        await super().create()
        if not self.app or (loop := self.app.loop) is None:
            raise ValueError("Cannot start TelnetServer without running loop")

        self.server = await loop.create_server(
            protocol_factory=lambda: telnetlib3.TelnetServer(**self.telnet_kwargs),
            host=self.host,
            port=self.port,
        )

    async def handle_connect(
        self, reader: telnetlib3.TelnetReader, writer: telnetlib3.TelnetWriter
    ):
        connection: TelnetConnection = self.connection_class(
            server=self, reader=reader, writer=writer
        )
        await self.connected(connection)



from __future__ import annotations

import logging
from collections import defaultdict
from collections.abc import Awaitable, Callable
from typing import TYPE_CHECKING, Any

from ..events import Event

# Type aliases
HandlerType = Callable[[Event], Awaitable[None]]
FilterType = dict[str, Any]
EventsType = dict[type[Event], list[tuple[HandlerType, FilterType]]]

if TYPE_CHECKING:
    from .app import App


logger = logging.getLogger("mara.event")


class EventManager:
    """
    Internal event manager

    Most apps will have one EventManager, created and managed by the App. In normal use
    use the ``App.listen()`` and ``App.trigger()`` methods instead of calling this
    directly.
    """

    app: App

    # Registered events
    events: EventsType

    # Defined event classes
    _known_events: dict[type[Event], None]

    def __init__(self, app: App):
        # Initialise events
        self.app = app
        self.events: EventsType = defaultdict(list)
        self._known_events: dict[type[Event], None] = {}

    def listen(
        self,
        event_class: type[Event],
        handler: HandlerType | None = None,
        **filters: FilterType,
    ):
        """
        Bind a handler callback to the specified event class, and to its subclasses

        Can be called directly::

            manager.listen(Event, handler)
            manager.listen(Event, handler)

        or can be called as a decorator with no handler argument::

            @manager.listen(Event)
            async def callback(event):
                ...

        Arguments:
            event_class (Type[Event]): The Event class to listen for
            handler (Awaitable | None): The handler, if not being decorated
            server (AbstractServer | List[AbstractServer] | None): The Server class or
                classes to filter inbound events
        """
        # Called directly
        if handler is not None:
            self._listen(event_class, handler, filters)
            return handler

        # Called as a decorator
        def decorator(fn):
            self._listen(event_class, fn, filters)
            return fn

        return decorator

    def _listen(
        self,
        event_class: type[Event],
        handler: HandlerType,
        filters: FilterType,
    ):
        """
        Internal method to recursively bind a handler to the specified event
        class and its subclasses. Call listen() instead.
        """
        # Recurse subclasses. Do it before registering for this event in case
        # they're not known yet, then they'll copy handlers for this event
        for subclass in event_class.__subclasses__():
            self._listen(subclass, handler, filters)

        # Register class
        self._ensure_known_event(event_class)
        self.events[event_class].append((handler, filters))

    def _ensure_known_event(self, event_class: type[Event]):
        """
        Ensure the event class is known to the service.

        If it is not, inherit handlers from its first base class
        """
        # If known, nothing to do
        if event_class in self._known_events:
            return

        # If base class isn't an event (ie we're dealing with Event), nothing to do
        base_cls = event_class.__bases__[0]
        if not issubclass(base_cls, Event):
            return

        # We'll know this going forwards, don't want to re-register
        self._known_events[event_class] = None

        # Ensure base class is known - if it isn't, we'll keep working up until we find
        # something we do know
        self._ensure_known_event(base_cls)

        # We've found a known event, propagate down any handlers

        # Propagating at registration means that we don't need to walk the MRO for
        # every event raised
        self.events[event_class] = self.events[base_cls][:]

    async def trigger(self, event: Event):
        """
        Trigger the specified event
        """
        # Make sure we've seen this event class before - will propagate handlers if not
        event_class = type(event)
        self._ensure_known_event(event_class)

        # Make sure all listeners have access to the app, in case they're out of scope
        event.app = self.app

        # Log the event
        logger.info(str(event))
        # self.app.log.event(event)

        # Call all handlers
        handler: HandlerType
        filters: FilterType
        for handler, filters in self.events[event_class]:
            # Catch stopped event
            if event.stopped:
                return

            # Filter anything which doesn't match
            # TODO: This could be enhanced to allow more complex filtering rules
            if not all(
                hasattr(event, filter_key)
                and getattr(event, filter_key) == filter_value
                for filter_key, filter_value in filters.items()
            ):
                return

            # Pass to the handler
            await handler(event)



import asyncio
import logging
import sys
from os import getenv


class Whitelist(logging.Filter):
    def __init__(self, *whitelist):
        self.whitelist = [logging.Filter(name) for name in whitelist]

    def filter(self, record):
        return any(f.filter(record) for f in self.whitelist)


def configure():
    """
    Configure logging

    Pick up the log level from the env var LOGLEVEL, otherwise default to INFO
    """
    # TODO: Simple configuration of what to log and where to log it to
    level_name = getenv("LOGLEVEL", "INFO")
    level = getattr(logging, level_name)
    logging.basicConfig(stream=sys.stdout, filemode="w", level=level)

    for handler in logging.root.handlers:
        handler.addFilter(Whitelist("mara", "tests"))


def get_tasks(loop):
    """
    List the current tasks in the loop

    Usage::

        logger.debug(get_tasks(app.loop))
    """
    tasks = asyncio.all_tasks(loop)
    return "Tasks: " + ", ".join(
        [f"{task.get_name()}: {task.get_coro().__name__}" for task in tasks]
    )



"""
Server events
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from .base import Event


__all__ = ["Server", "ListenStart", "Suspend", "ListenStop"]


if TYPE_CHECKING:
    from ..servers import AbstractServer


class Server(Event):
    "Server event"
    server: AbstractServer

    def __init__(self, server: AbstractServer):
        self.server = server

    def __str__(self) -> str:
        return f"{super().__str__()}: {self.server}"


class ListenStart(Server):
    "Server listening"


class Suspend(Server):
    "Server has been suspended"


class ListenStop(Server):
    "Server is no longer listening"



from __future__ import annotations

from .app import (  # noqa
    App,
    PostRestart,
    PostStart,
    PostStop,
    PreRestart,
    PreStart,
    PreStop,
)
from .base import Event  # noqa
from .connection import Connect, Connection, Disconnect, Receive  # noqa
from .server import ListenStart, ListenStop, Server, Suspend  # noqa



from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, Any, Coroutine, List

from ..events import Event, PostStart, PostStop, PreStart, PreStop
from ..status import Status
from . import event_manager
from .logging import configure as configure_logging

if TYPE_CHECKING:
    from ..servers import AbstractServer
    from ..timers import AbstractTimer

configure_logging()
logger = logging.getLogger("mara.app")


class App:
    """
    Orchestrate servers and tasks
    """

    loop: asyncio.AbstractEventLoop | None = None
    servers: List[AbstractServer]
    events: event_manager.EventManager
    timers: List[AbstractTimer]
    _status: Status = Status.IDLE

    def __init__(self):
        self.servers = []
        self.timers = []

        self.events = event_manager.EventManager(self)

    def add_server(self, server: AbstractServer) -> AbstractServer:
        """
        Add a new Server instance to the async loop

        The server will start listening when the app is ``run()``. If it is already
        running, it will start listening immediately.
        """
        logger.debug(f"Add server {server}")
        self.servers.append(server)

        if self.loop:
            logger.debug(f"Running server {server}")
            self.create_task(server.run(self))

        return server

    def remove_server(self, server: AbstractServer):
        """
        Stop and remove the specified Server instance
        """
        if server.status == Status.RUNNING:
            logger.debug(f"Stopping server {server}")
            server.stop()

        logger.debug(f"Removing server {server}")
        self.servers.remove(server)

    def add_timer(self, timer: AbstractTimer) -> AbstractTimer:
        """
        Add a new Timer instance to the async loop

        The timer will start when the app is ``run()``. If it is already running, it
        will start immediately.

        Returns the timer instance. Because timer instances are callable decorators, you
        can use this as a decorator to define a timer and its function in one::

            @app.add_timer(PeriodicTimer(every=1))
            async def tick(timer):
                ...

        which is shorthand for::

            async def tick(timer):
                ...

            timer = PeriodicTimer(every=60)
            app.add_timer(timer)
            timer(tick)
        """
        logger.debug(f"Add timer {timer}")
        self.timers.append(timer)

        if self.loop:
            logger.debug(f"Timer {timer} starting")
            self.create_task(timer.run(self))
            logger.debug(f"Timer {timer} stopped")

            # TODO: extend self.create_task to callback to a fn to clean up self.timers

        return timer

    def run(self, debug=True):
        """
        Start the main app async loop

        This will start any Servers which have been added with ``add_server()``
        """
        self._status = Status.STARTING

        # TODO: Should add some more logic around here from asyncio.run
        self.loop = loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        logger.debug("Loop starting")
        loop.run_until_complete(self.events.trigger(PreStart()))

        for server in self.servers:
            self.create_task(server.run(self))

        for timer in self.timers:
            self.create_task(timer.run(self))

        logger.debug("Loop running")
        self._status = Status.RUNNING
        loop.run_until_complete(self.events.trigger(PostStart()))

        try:
            loop.run_forever()
        finally:
            logger.debug("Loop stopping")
            self._status = Status.STOPPING
            loop.run_until_complete(self.events.trigger(PreStop()))
            loop.run_until_complete(loop.shutdown_asyncgens())
            loop.close()
            self._status = Status.STOPPED
            self.loop = None
            logger.debug("Loop stopped")

            # Loop has terminated
            asyncio.run(self.events.trigger(PostStop()))

    def create_task(self, task_fn: Coroutine[Any, Any, Any]):
        if self.loop is None:
            # TODO: Handle pending tasks here
            raise ValueError("Loop is not running")
        task = self.loop.create_task(task_fn)
        task.add_done_callback(self._handle_task_complete)
        return task

    def _handle_task_complete(self, task: asyncio.Task):
        try:
            task.result()
        except asyncio.CancelledError:
            pass
        except Exception:
            logger.exception("Task failed")

    def on(
        self,
        event_class: type[Event],
        handler: event_manager.HandlerType | None = None,
        **filters: event_manager.FilterType,
    ):
        """
        Bind a handler callback to the specified event class, and to its subclasses

        Arguments:

            event_class (Type[Event]): The Event class to listen for
            handler (Awaitable | None): The handler, if not being decorated
            **filters: Key value pairs to match against inbound events

        Can be called directly::

            app.on(Event, handler)

        or can be called as a decorator with no handler argument::

            @app.on(Event)
            async def callback(event):
                ...

        and can filter based on event attributes

            @app.on(ServerEvent, server=telnet)
            async def callback(event):
                ...
        """
        return self.events.listen(event_class, handler, **filters)

    @property
    def status(self):
        """
        Get the status of the app and its servers

        Difference between this and self._status is when the app loop is RUNNING, this
        will return STARTING until all servers are listening
        """
        if not self._status == Status.RUNNING:
            return self._status

        if all([server.status == Status.RUNNING for server in self.servers]):
            return Status.RUNNING
        return Status.STARTING

    def stop(self):
        """
        This will ask the main loop to stop, shutting down all servers, connections and
        other async tasks.
        """
        if self.loop:
            logger.debug("Stopping servers")
            for server in self.servers:
                server.stop()

            logger.debug("Requesting loop stop")
            self.loop.stop()



"""
Service events
"""
from __future__ import annotations

from .base import Event


__all__ = [
    "App",
    "PreStart",
    "PostStart",
    "PreStop",
    "PostStop",
    "PreRestart",
    "PostRestart",
]


class App(Event):
    "Service event"


class PreStart(App):
    "Service starting"


class PostStart(App):
    "Service started"


class PreStop(App):
    "Service stopping"


class PostStop(App):
    "Service stopped"


class PreRestart(App):
    "Service restarting"


class PostRestart(App):
    "Service restarted"



import json
from typing import Any

from .base import Store, store_classes


class DictStore(Store, dict):
    """
    A dict-based store with no permanence

    For convenience, keys are also available as attributes.

    To freeze a special case value, define:

    * ``freeze_KEY(key:str, value:Any) -> Any`` which must return a value serialisable
      by ``json``, and
    * ``thaw_KEY(key:str, value:Any) -> Any`` which converts the frozen value back into
      the real value

    replacing ``KEY`` with the key of the special case value.
    """

    def __getattr__(self, key):
        return self[key]

    def __setattr__(self, key, value):
        self[key] = value

    async def store(self) -> str:
        # Start by making all values safe to serialise
        safe = {}
        for key, value in self.items():
            safe[key] = await self._freeze_item(key, value)

        # Now serialise to json
        data = json.dumps(safe)
        return data

    async def _freeze_item(self, key: str, value: Store | Any) -> Any:
        """
        Make value safe to serialise
        """
        freezer_fn_name = f"freeze_{key}"
        if freezer_fn_name in dir(self):
            # can't use hasattr/getattr because of __getattr__
            freezer_fn = getattr(self, freezer_fn_name)
            return freezer_fn(key, value)

        elif isinstance(value, Store):
            data = await value.store()
            return {
                "_store_class": type(value).__name__,
                "data": data,
            }

        return value

    @classmethod
    async def restore(cls, data: str):
        """
        Deserialise and create a new object
        """
        # Deserialise from JSON
        safe = json.loads(data)

        # Thaw values
        kwargs = {}
        for key, value in safe.items():
            kwargs[key] = await cls._thaw_item(key, value)

        return cls(**kwargs)

    @classmethod
    async def _thaw_item(cls, key: str, value: Any) -> Any:
        thaw_fn_name = f"freeze_{key}"
        if thaw_fn_name in dir(cls):
            # can't use hasattr/getattr because of __getattr__
            thaw_fn = getattr(cls, thaw_fn_name)
            return thaw_fn(key, value)

        elif isinstance(value, dict) and "_store_class" in value:
            store_cls_name = value["_store_class"]
            store_cls = store_classes[store_cls_name]
            obj = await store_cls.restore(value["data"])
            return obj

        return value



from __future__ import annotations

import asyncio
import logging
from time import time
from typing import TYPE_CHECKING, Awaitable, Callable


if TYPE_CHECKING:
    from .. import App


logger = logging.getLogger("mara.timer")


class AbstractTimer:
    app: App
    callback: Callable[[AbstractTimer], Awaitable[None]] | None = None
    running: bool = False

    def __str__(self):
        if self.callback is None:
            return str(id(self))
        return self.callback.__name__

    def __call__(
        self, callback: Callable[[AbstractTimer], Awaitable[None]]
    ) -> Callable[[AbstractTimer], Awaitable[None]]:
        logger.debug(f"Timer {self} assigned callback {callback.__name__}")
        self.callback = callback
        return callback

    async def run(self, app: App):
        self.app = app

        logger.debug(f"Timer {self} starting")
        if not self.callback:
            raise ValueError(f"Timer {self} has not been assigned a callback")

        self.running = True
        last_due = time()
        while self.running:
            now = time()
            next_due = self.next_due(last_due, now)
            if not next_due:
                logger.debug(f"Timer {self} at {now} has no next due, stopping")
                break

            logger.debug(f"Timer {self} at {now} is next due {next_due}")
            await asyncio.sleep(next_due - now)
            logger.debug(f"Timer {self} active")
            await self.callback(self)

        self.running = False
        logger.debug(f"Timer {self} stopped")
        # TODO
        # self.app.remove_timer(self)

    def next_due(self, last_due: float, now: float) -> int | float | None:
        """
        Return unix time for the next time the trigger is due

        Arguments:

            last_due (float): when this was last due
            now (float): when this was actually triggered

        Timers using ``now`` for relative calculations will result in drift. Timers
        should ensure that their next due is after ``now``.

        A next due time of 0 or None will terminate the timer
        """
        return 0

    def stop(self):
        logger.debug(f"Timer {self} stopping")
        self.running = False



from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, Generic, TypeVar

from ...events import Connect, Disconnect, Receive
from ...storage.dict import DictStore

if TYPE_CHECKING:
    from ..base import AbstractServer


ContentType = TypeVar("ContentType")

logger = logging.getLogger("mara.connection")


class AbstractConnection(Generic[ContentType]):
    server: AbstractServer
    connected: bool
    read_task: asyncio.Task
    write_task: asyncio.Task
    write_queue: asyncio.Queue
    session: DictStore

    def __init__(self, server: AbstractServer):
        self.server = server
        self.connected = True
        self.session = DictStore()
        # TODO: Queue(maxsize=?) - configure from server
        self.write_queue = asyncio.Queue()

    def __str__(self):
        return "unknown"

    async def read(self) -> ContentType:
        raise NotImplementedError()

    def write(self, data: ContentType):
        """
        Write to the outbound queue
        """
        self.write_queue.put_nowait(data)

    async def flush(self):
        """
        Wait for all outbound data to be sent
        """
        await self.write_queue.join()

    async def _write(self, data: ContentType):
        """
        Write to the connection
        """
        raise NotImplementedError()

    async def close(self):
        logger.info(f"Connection {self} closed")
        await self.server.disconnected(self)

    def run(self):
        """
        Add the connection read and write tasks to the app's loop
        """
        if not self.server.app:
            raise ValueError("Connection read loop must be part of an active server")
        self.read_task = self.server.app.create_task(self._read_loop())
        self.write_task = self.server.app.create_task(self._write_loop())

    async def _read_loop(self):
        app = self.server.app
        if not app:
            raise ValueError("Connection read loop must be part of an active server")
        await app.events.trigger(Connect(self))
        logger.info(f"Connection {self} connected")
        while self.connected:
            data: ContentType = await self.read()
            if data:
                await app.events.trigger(Receive(self, data))

        logger.info(f"Connection {self} disconnected")
        await app.events.trigger(Disconnect(self))

    async def _write_loop(self):
        while self.connected:
            data: ContentType = await self.write_queue.get()
            await self._write(data)
            self.write_queue.task_done()



from aiohttp import WSMsgType
from aiohttp.web import Request, WebSocketResponse

from .base import AbstractConnection


class WebSocketConnection(AbstractConnection[str]):
    request: Request
    ws: WebSocketResponse

    def __str__(self):
        return str(self.request.remote)

    async def prepare(self, request: Request):
        self.request = request
        self.ws = WebSocketResponse()
        await self.ws.prepare(request)

    async def read(self) -> str:
        msg = await self.ws.receive()
        if msg.type == WSMsgType.TEXT:
            return msg.data
        elif msg.type == WSMsgType.ERROR:
            self.connected = False
            return ""
        # if msg.type in (WSMsgType.CLOSE, WSMsgType.CLOSING, WSMsgType.CLOSED):
        #    raise StopAsyncIteration
        return ""

    async def _write(self, data: str):
        await self.ws.send_str(data)



from __future__ import annotations

import asyncio
import logging
from pathlib import Path
from ssl import Purpose, SSLContext, create_default_context

from ..constants import DEFAULT_HOST, DEFAULT_PORT
from .base import AbstractAsyncioServer
from .connections.socket import SocketConnection, SocketMixin, TextConnection

logger = logging.getLogger("mara.server")


class AbstractSocketServer(AbstractAsyncioServer):
    connection_class: type[SocketMixin]
    _host: str
    _port: int
    _ssl_context: SSLContext | None = None

    def __init__(
        self,
        host: str = DEFAULT_HOST,
        port: int = DEFAULT_PORT,
        ssl: SSLContext | None = None,
        ssl_cert: str | Path | None = None,
        ssl_key: str | Path | None = None,
    ):
        self.host = host
        self.port = port
        if ssl or ssl_cert:
            if ssl and ssl_cert:
                raise ValueError("Cannot provide both an SSL context and certificate")

            if ssl_cert:
                ssl = create_default_context(Purpose.CLIENT_AUTH)
                ssl.check_hostname = False
                ssl.load_cert_chain(certfile=ssl_cert, keyfile=ssl_key)

            if not isinstance(ssl, SSLContext):
                raise ValueError("ssl must be an ssl.SSLContext")

            self._ssl_context = ssl

        super().__init__()

    def __str__(self):
        return f"{'SSL 'if self._ssl_context else ''}Socket {self.host}:{self.port}"

    async def create(self):
        await super().create()

        self.server = await asyncio.start_server(
            client_connected_cb=self.handle_connect,
            host=self.host,
            port=self.port,
            ssl=self._ssl_context,
        )

    async def handle_connect(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ):
        connection: SocketMixin = self.connection_class(
            server=self, reader=reader, writer=writer
        )
        await self.connected(connection)


class SocketServer(AbstractSocketServer):
    connection_class: type[SocketConnection] = SocketConnection


class TextServer(AbstractSocketServer):
    connection_class: type[TextConnection] = TextConnection



store_classes = {}


class Store:
    """
    A class which can store and restore its data

    Each Store subclass is responsible for how it stores and restores itself using a
    single string value; for example, the DictStore serialises to json which is
    returned raw, while the SqliteStore saves data to an sqlite database and returns the
    database ID of the object.

    Each Store subclass must have a unique name, as it uses that to register itself with
    the global store class registry, so that store class names mentioned in serialised
    data can be restored.

    A subclass must implement store() and restore()

    Abstract subclasses should set abstract=True in their definition, eg::

        class MyAbstractStore(Store, abstract=True):
            ...
    """

    def __init_subclass__(cls, *, abstract=False, **kwargs):
        super().__init_subclass__(**kwargs)
        if not abstract:
            store_classes[cls.__name__] = cls

    async def store(self) -> str:
        raise NotImplementedError()

    @classmethod
    async def restore(cls, data: str) -> None:
        raise NotImplementedError()



from __future__ import annotations

from typing import TYPE_CHECKING

from telnetlib3 import TelnetReader, TelnetWriter

from .base import AbstractConnection

if TYPE_CHECKING:
    from ..servers.telnet import TelnetServer


class TelnetConnection(AbstractConnection[str]):
    reader: TelnetReader
    writer: TelnetWriter
    _str: str | None = None

    def __init__(
        self,
        server: TelnetServer,
        reader: TelnetReader,
        writer: TelnetWriter,
    ):
        super().__init__(server)
        self.reader = reader
        self.writer = writer
        self._str = None

    def __str__(self) -> str:
        # Cache value, it won't be available after connection closes
        if not self._str:
            ip, port = self.writer.get_extra_info("peername")
            self._str = str(ip)
        return self._str

    async def _write(self, data: str):
        # TODO: says it takes bytes but needs str
        self.writer.write(data)  # type: ignore
        await self.writer.drain()
        self._check_is_active()

    def _check_is_active(self):
        if self.reader.at_eof() or self.writer.transport.is_closing():
            self.connected = False

    async def close(self):
        # Close the streams
        self.writer.close()

        # Can't wait to make sure the socket is closed:
        # https://github.com/jquast/telnetlib3/issues/55
        #
        # await self.writer.wait_closed()

        # Terminate the listener loop
        # TODO
        # this.listener_task

        await super().close()

    async def read(self) -> str:
        # TODO: read size and buffers
        data = await self.reader.readline()
        data = data.rstrip("\r\n")
        self._check_is_active()
        return data

    def write(self, data: str, *, end: str = "\r\n"):
        data = f"{data}{end}"
        super().write(data)



"""
Mara events
"""
from __future__ import annotations

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from ..app import App


class Event(object):
    """
    Non-specific event

    All events are derived from this class.
    """

    app: App | None

    def __init__(self):
        self.stopped = False

        # App will be added by event handler
        self.app = None

    def stop(self):
        """
        Stop the event from being passed to any more handlers
        """
        self.stopped = True

    def __str__(self):
        """
        Return this event as a string
        """
        label = getattr(self.__class__, "hint", self.__class__.__doc__ or "")
        label = (label.strip().splitlines()[0],)
        return f"[{self.__class__.__name__}]: {label}"



from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING

from ..app import event_manager
from ..events import Event, ListenStart, ListenStop
from ..status import Status

if TYPE_CHECKING:
    from ..app import App
    from .connections import AbstractConnection

logger = logging.getLogger("mara.server")

DeferredEventType = tuple[
    type[Event], event_manager.HandlerType, event_manager.FilterType
]


class AbstractServer:
    app: App | None = None
    connections: list[AbstractConnection]
    _status: Status = Status.IDLE
    _events: list[DeferredEventType]

    def __init__(self):
        self.connections = []
        self._events: list[DeferredEventType] = []

    def __str__(self):
        return self.__class__.__name__

    async def run(self, app: App):
        """
        Create the server
        """
        self.app = app

        # Register deferred events
        for event_class, handler, filters in self._events:
            self.app.on(event_class=event_class, handler=handler, **filters)

        # Initialise server and listen
        await self.create()
        await self.listen()

    async def create(self):
        logger.debug(f"Server starting: {self}")
        self._status = Status.STARTING

    async def listen(self):
        """
        Main listening loop
        """
        logger.debug(f"Server running: {self}")
        self._status = Status.RUNNING

        # Raise the event
        await self.app.events.trigger(ListenStart(self))

        # Start the server
        await self.listen_loop()

        # Look has exited, likely due to a call to self.stop()
        logger.debug(f"Server stopping: {self}")
        await self.app.events.trigger(ListenStop(self))
        logger.info(f"Server stopped: {self}")
        self._status = Status.STOPPED

    async def listen_loop(self):
        """
        Placeholder for subclasses to implement their listen loop
        """
        pass

    async def connected(self, connection: AbstractConnection):
        """
        Register a new client connection and start the lifecycle
        """
        logger.info(f"[{self}] Connection from {connection}")
        self.connections.append(connection)
        connection.run()

    async def disconnected(self, connection: AbstractConnection):
        """
        Unregister a connection that has disconnected
        """
        self.connections.remove(connection)

    def stop(self):
        """
        Shut down the server
        """
        self._status = Status.STOPPING
        logger.info(f"Server closing: {self}")

    @property
    def status(self) -> Status:
        return self._status

    def on(
        self,
        event_class: type[Event],
        handler: event_manager.HandlerType | None = None,
        **filters: event_manager.FilterType,
    ):
        filters["server"] = self

        if self.app is not None:
            return self.app.on(event_class=event_class, handler=handler, **filters)

        def defer(handler):
            self._events.append((event_class, handler, filters))
            return handler

        if handler is not None:
            defer(handler)
            return handler

        return defer


class AbstractAsyncioServer(AbstractServer):
    """
    Base class for servers based on asyncio.Server
    """

    server: asyncio.base_events.Server

    async def listen_loop(self):
        async with self.server:
            await self.server.serve_forever()

    def stop(self):
        """
        Shut down the server
        """
        super().stop()
        self.server.close()

    @property
    def status(self) -> Status:
        status = super().status

        if status == Status.RUNNING and not self.server.is_serving():
            return Status.STARTING

        return status



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Protocol

from .base import AbstractConnection

if TYPE_CHECKING:
    from ..servers import AbstractServer


class ConnectionCanStream(Protocol):
    reader: asyncio.StreamReader
    writer: asyncio.StreamWriter

    def __init__(
        self,
        server: AbstractServer,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
    ): ...


class SocketMixin(AbstractConnection):
    """
    Mixin for any connection class which uses byte sockets
    """

    reader: asyncio.StreamReader
    writer: asyncio.StreamWriter

    def __init__(
        self,
        server: AbstractServer,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
    ):
        super().__init__(server)
        self.reader = reader
        self.writer = writer

    def __str__(self) -> str:
        ip, port = self.writer.get_extra_info("peername")
        return str(ip)

    async def _write(self, data: bytes):
        self.writer.write(data)
        await self.writer.drain()
        self._check_is_active()

    def _check_is_active(self):
        if self.reader.at_eof() or self.writer.transport.is_closing():
            self.connected = False

    async def close(self):
        # Close the streams
        self.writer.close()
        await self.writer.wait_closed()

        # Terminate the listener loop
        # TODO
        # this.listener_task

        await super().close()


class SocketConnection(SocketMixin, AbstractConnection[bytes]):
    """
    Read and write bytes
    """

    async def read(self) -> bytes:
        # TODO: read size and buffers
        data = await self.reader.read(1024)
        self._check_is_active()
        return data


class TextConnection(SocketMixin, AbstractConnection[str]):
    """
    Read and write unicode over an underlying byte socket
    """

    async def read(self) -> str:
        # TODO: read size and buffers
        try:
            data: bytes = await self.reader.readuntil(b"\r\n")
        except asyncio.exceptions.IncompleteReadError:
            self.connected = False
            return ""
        data = data.rstrip(b"\r\n")
        self._check_is_active()
        text: str = data.decode()
        return text

    def write(self, data: str, *, end: str = "\r\n"):
        raw_data: bytes = f"{data}{end}".encode()
        super().write(raw_data)
